from abc import ABC, abstractmethod
import json
import pickle
from typing import Any, Dict, List, Union

VERSION = 3


def ensure_size(text: Any, max_length: int = 300) -> str:
    """
    Ensure a text size is <= max_length. If exceeds, return a truncated version ending with '..'.
    Always coerces input to string safely.
    """
    if text is None:
        return ""
    s = str(text)
    if len(s) <= max_length:
        return s
    return s[: max_length - 2] + ".."


class Parcel(ABC):
    def __init__(self, content: Any = None, topic_return: Any = None):
        self.version: int = VERSION
        self.content: Any = content
        self.topic_return: Any = topic_return
        self.error: Any = None

    def __str__(self) -> str:
        def _convert_content(content: Any) -> Union[str, Dict[str, Any], List[Any]]:
            """Convert parcel content recursively to a JSON-serializable, readable structure."""
            # bytes -> utf-8 text (replace undecodable), keep length info on failure
            if isinstance(content, (bytes, bytearray)):
                try:
                    return ensure_size(bytes(content).decode("utf-8", "replace"))
                except Exception as e:  # very defensive; decode with 'replace' should not fail
                    return f"len(<binary data>): {len(content)}, error: {e}"

            # dict -> recurse values
            if isinstance(content, dict):
                return {str(k): _convert_content(v) for k, v in content.items()}

            # list/tuple/set -> list of converted
            if isinstance(content, (list, tuple, set)):
                return [_convert_content(v) for v in content]

            # everything else -> string
            try:
                return ensure_size(str(content))
            except Exception:
                return "<unstringifiable>"

        body = {
            "version": self.version,
            "content": _convert_content(self.content),
            "topic_return": self.topic_return,
            "error": self.error,
        }

        # Ensure JSON serializable (final safety net)
        try:
            return json.dumps(body, indent=4, ensure_ascii=False)
        except TypeError:
            def _force(v: Any):
                if isinstance(v, (str, int, float, bool)) or v is None:
                    return v
                if isinstance(v, dict):
                    return {str(k): _force(val) for k, val in v.items()}
                if isinstance(v, list):
                    return [_force(i) for i in v]
                return str(v)

            safe_body = {k: _force(v) for k, v in body.items()}
            return json.dumps(safe_body, indent=4, ensure_ascii=False)

    @staticmethod
    def from_content(content: Any) -> "Parcel":
        is_binary = isinstance(content, (bytes, bytearray))
        if is_binary:
            return BinaryParcel(content)
        return TextParcel(content)

    @staticmethod
    def from_payload(payload: bytes) -> "Parcel":
        if BinaryParcel.is_payload(payload):
            return BinaryParcel.from_payload(payload)
        if TextParcel.is_payload(payload):
            return TextParcel.from_payload(payload)
        raise TypeError("Not valid Parcel payload.")

    @staticmethod
    def is_payload(payload: bytes) -> bool:
        return BinaryParcel.is_payload(payload) or TextParcel.is_payload(payload)

    def _get_managed_data(self) -> Dict[str, Any]:
        return {
            "version": VERSION,
            "content": self.content,
            "topic_return": self.topic_return,
            "error": self.error,
        }

    def _set_managed_data(self, managed_data: Dict[str, Any]) -> None:
        self.version = managed_data.get("version", VERSION)
        self.content = managed_data.get("content")
        self.topic_return = managed_data.get("topic_return")
        self.error = managed_data.get("error")

    @abstractmethod
    def payload(self) -> bytes:
        """Return the serialized bytes payload."""
        raise NotImplementedError

    # Subscript operations for dict-like content
    def __getitem__(self, key: Any) -> Any:
        return self.get(key)

    def __setitem__(self, key: Any, value: Any) -> None:
        self.set(key, value)

    def get(self, key: Any, default: Any = None) -> Any:
        if isinstance(self.content, dict):
            return self.content.get(key, default)
        raise TypeError("self.content is not a dictionary. Get operation is not allowed.")

    def set(self, key: Any, value: Any) -> None:
        if self.content is None:
            self.content = {}
        if isinstance(self.content, dict):
            self.content[key] = value
        else:
            raise TypeError("self.content is not a dictionary. Set operation is not allowed.")


class BinaryParcel(Parcel):
    HEAD = b"application/pickle|"

    def __init__(self, content: Any = None, topic_return: Any = None):
        super().__init__(content, topic_return)

    @staticmethod
    def from_payload(payload: bytes) -> "BinaryParcel":
        if not isinstance(payload, (bytes, bytearray)):
            raise TypeError("payload must be bytes for BinaryParcel")
        raw = bytes(payload)[len(BinaryParcel.HEAD):]
        managed = pickle.loads(raw)
        pcl = BinaryParcel()
        pcl._set_managed_data(managed)
        return pcl

    @staticmethod
    def is_payload(payload: bytes) -> bool:
        return bool(payload) and isinstance(payload, (bytes, bytearray)) and bytes(payload).startswith(BinaryParcel.HEAD)

    def payload(self) -> bytes:
        return BinaryParcel.HEAD + pickle.dumps(self._get_managed_data(), protocol=pickle.HIGHEST_PROTOCOL)


class TextParcel(Parcel):
    HEAD = b"text/json|"

    def __init__(self, content: Any = None, topic_return: Any = None):
        super().__init__(content, topic_return)

    @staticmethod
    def from_payload(payload: bytes) -> "TextParcel":
        if not isinstance(payload, (bytes, bytearray)):
            raise TypeError("payload must be bytes for TextParcel")
        raw = bytes(payload)[len(TextParcel.HEAD):]
        data = json.loads(raw.decode("utf-8", "replace"))
        pcl = TextParcel()
        pcl._set_managed_data(data)
        return pcl

    @staticmethod
    def is_payload(payload: bytes) -> bool:
        return bool(payload) and isinstance(payload, (bytes, bytearray)) and bytes(payload).startswith(TextParcel.HEAD)

    def payload(self) -> bytes:
        data = json.dumps(self._get_managed_data(), ensure_ascii=False).encode("utf-8")
        return TextParcel.HEAD + data



from paho.mqtt.client import Client
from paho.mqtt.enums import CallbackAPIVersion
import logging, os, threading
logger = logging.getLogger(os.getenv('LOGGER_NAME'))

from .message_broker import MessageBroker
from .notifier import BrokerNotifier


class MqttBroker(MessageBroker):
    def __init__(self, notifier: BrokerNotifier, *, wait: bool = True, timeout: float = 10.0):
        """建立 MQTT Broker 包裝類別
        :param notifier: BrokerNotifier 實例
        :param wait: 是否等待連線完成才返回 start()
        :param timeout: 最長等待秒數
        """
        self._client = Client(callback_api_version=CallbackAPIVersion.VERSION2,
                              reconnect_on_failure=False)
        self.host = ""
        self.port = 0
        self.keepalive = 0

        # 連線事件控制
        self._connected_evt = threading.Event()
        self._connect_ok = False
        self._connect_err = None

        # 等待連線的行為設定
        self._wait = wait
        self._timeout = timeout

        logger.info(f"MQTT broker initialized with notifier: {notifier}, wait={wait}, timeout={timeout}")
        super().__init__(notifier=notifier)

    def _on_connect(self, client, userdata, flags, reasonCode, properties):
        if reasonCode == 0:
            self._connect_ok = True
            self._connect_err = None
            logger.info(f"MQTT broker connected: {self.host}:{self.port}, keepalive={self.keepalive}")
            try:
                self._notifier._on_connect()
            finally:
                self._connected_evt.set()
        else:
            self._connect_ok = False
            name = getattr(reasonCode, "getName", lambda: str(reasonCode))()
            self._connect_err = f"{reasonCode} ({name})"
            logger.error(f"MQTT connection failed: {self._connect_err}")
            self._connected_evt.set()


    def _on_disconnect(self, client, userdata, _flags, reasonCode, _properties):
        logger.warning(f"MQTT disconnected: {reasonCode}")


    def _on_message(self, client, db, message):
        try:
            self._notifier._on_message(message.topic, message.payload)
        except Exception as ex:
            logger.exception(ex)


    def start(self, options: dict):
        logger.info("MQTT broker is starting...")

        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect
        self._client.on_message = self._on_message

        self.host = options.get("host", "localhost")
        self.port = int(options.get("port", 1883))
        self.keepalive = int(options.get("keepalive", 60))

        if username := options.get("username"):
            self._client.username_pw_set(username, options.get("password"))

        # 初始化事件
        self._connected_evt.clear()
        self._connect_ok = False
        self._connect_err = None

        self._client.connect(self.host, self.port, self.keepalive)
        self._client.loop_start()

        if not self._wait:
            return True

        if not self._connected_evt.wait(self._timeout):
            logger.error(f"MQTT connect timeout after {self._timeout}s")
            self._client.loop_stop()
            self._client.disconnect()
            raise TimeoutError(f"MQTT connect timeout ({self.host}:{self.port})")

        if not self._connect_ok:
            self._client.loop_stop()
            self._client.disconnect()
            raise ConnectionError(f"MQTT connect failed: {self._connect_err}")

        return True

    def stop(self):
        logger.warning("MQTT broker is stopping...")
        self._client.disconnect()
        self._client.loop_stop()

    def publish(self, topic: str, payload):
        return self._client.publish(topic=topic, payload=payload)

    def subscribe(self, topic: str, data_type):
        return self._client.subscribe(topic=topic)
